# -*- coding: utf-8 -*-
# @Author: Theo Lemaire
# @Date:   2022-08-15 10:21:10
# @Last Modified by:   Theo Lemaire
# @Last Modified time: 2023-05-17 10:54:10

''' List all available VISA resources '''

from instrulink import list_visa_resources

list_visa_resources()
